/*
    PPDocument_Layers.m

    Copyright 2013-2016 Josh Freeman
    http://www.twilightedge.com

    This file is part of PikoPixel for Mac OS X and GNUstep.
    PikoPixel is a graphical application for drawing & editing pixel-art images.

    PikoPixel is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License version 3,
    as published by the Free Software Foundation.

    PikoPixel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#import "PPDocument.h"

#import "PPDocument_Notifications.h"
#import "PPDocumentLayer.h"
#import "PPGeometry.h"
#import "NSImage_PPUtilities.h"
#import "NSBitmapImageRep_PPUtilities.h"


@interface PPDocument (LayersPrivateMethods)

- (bool) hasLayerAtIndex: (int) index;
- (bool) isValidLayerInsertionIndex: (int) index;
- (int) indexOfLayer: (PPDocumentLayer *) layer;

- (void) invalidateAllRelativeCachedLayerImagesForIndex: (int) index;
- (void) invalidateCachedLayerImagesAtIndex: (int) index;
- (void) invalidateImagesInCache: (NSImage **) cachedImagesArray
            fromIndex: (int) startIndex
            toIndex: (int) endIndex;
- (void) removeAllCachedLayersImages;

- (NSImage *) cachedOverlayersImageForIndex: (int) index;
- (NSImage *) cachedUnderlayersImageForIndex: (int) index;

- (NSImage *) mergedLayersImageFromIndex: (int) firstIndex
                toIndex: (int) lastIndex;

- (NSBitmapImageRep *) mergedLayersBitmapFromIndex: (int) firstIndex
                        toIndex: (int) lastIndex;

- (void) updateMergedVisibleLayersBitmapInRect: (NSRect) rect
            indexOfUpdatedLayer: (int) indexOfUpdatedLayer;

- (NSString *) uniqueLayerNameWithRoot: (NSString *) rootName;
- (NSString *) duplicateNameForLayerName: (NSString *) layerName;

- (void) setupDrawingLayerWithLayerAtIndex: (int) newDrawingLayerIndex
            andPostNotification: (bool) shouldPostNotification;

- (void) setupDrawingLayerWithLayerAtIndex: (int) newDrawingLayerIndex;

- (void) setupDrawingLayerMembers;

- (bool) handleUpdateToLayer: (PPDocumentLayer *) layer;
- (void) handleUpdateToDrawingLayerInRect: (NSRect) updateRect;

- (void) updateDissolvedDrawingLayerBitmapInRect: (NSRect) updateRect;

- (void) insertArchivedLayer: (NSData *) archivedLayer
            atIndex: (int) index
            andSetAsDrawingLayer: (bool) shouldSetAsDrawingLayer;

- (bool) setLayersWithArchivedLayersData: (NSData *) archivedLayersData;

- (void) setName: (NSString *) name forLayerAtIndex: (int) index;
- (void) setEnabledFlag: (bool) isEnabled forLayerAtIndex: (int) index;
- (void) setOpacity: (float) opacity forLayerAtIndex: (int) index;

- (void) copyTIFFData: (NSData *) tiffData
            toLayerAtIndex: (int) index
            atPoint: (NSPoint) origin;

- (void) recacheMergedVisibleLayersThumbnailImageInBounds: (NSRect) bounds;
- (void) recacheDissolvedDrawingLayerThumbnailImageInBounds: (NSRect) bounds;

@end

@implementation PPDocument (Layers)

- (int) numLayers
{
    return _numLayers;
}

- (PPDocumentLayer *) layerAtIndex: (int) index
{
    if (![self hasLayerAtIndex: index])
    {
        return nil;
    }

    return (PPDocumentLayer *) [_layers objectAtIndex: index];
}

- (void) createNewLayer
{
    NSString *newLayerName;
    PPDocumentLayer *newLayer;
    int insertionIndex;

    if (!_numLayers)
    {
        newLayerName = @"Main Layer";
        insertionIndex = 0;
    }
    else
    {
        newLayerName = [self uniqueLayerNameWithRoot: @"New Layer"];
        insertionIndex = _indexOfDrawingLayer + 1;
    }

    newLayer = [PPDocumentLayer layerWithSize: _canvasFrame.size andName: newLayerName];

    if (!newLayer)
        return;

    [self insertLayer: newLayer atIndex: insertionIndex andSetAsDrawingLayer: YES];

    [[self undoManager] setActionName: @"Add Layer"];
}

- (void) insertLayer: (PPDocumentLayer *) layer
            atIndex: (int) index
            andSetAsDrawingLayer: (bool) shouldSetAsDrawingLayer
{
    int oldDrawingLayerIndex;
    NSUndoManager *undoManager;

    if (!layer || ![self isValidLayerInsertionIndex: index])
    {
        return;
    }

    // manually invalidate relevant cached images that won't be invalidated later by
    // invalidateAllRelativeCachedLayerImagesForIndex: (called by handleUpdateToLayerAtIndex:)

    [self invalidateImagesInCache: _cachedOverlayImages
                        fromIndex: index
                        toIndex: _numLayers - 1];

    [layer setDelegate: self];

    [_layers insertObject: layer atIndex: index];
    _numLayers = [_layers count];

    oldDrawingLayerIndex = _indexOfDrawingLayer;

    if (shouldSetAsDrawingLayer)
    {
        [self setupDrawingLayerWithLayerAtIndex: index andPostNotification: NO];
    }
    else if (_indexOfDrawingLayer >= index)
    {
        _indexOfDrawingLayer++;
    }

    [self handleUpdateToLayerAtIndex: index inRect: _canvasFrame];

    undoManager = [self undoManager];

    if (shouldSetAsDrawingLayer)
    {
        [[undoManager prepareWithInvocationTarget: self]
                                    setupDrawingLayerWithLayerAtIndex: oldDrawingLayerIndex];
    }

    [[undoManager prepareWithInvocationTarget: self] removeLayerAtIndex: index];

    if (![undoManager isUndoing] && ![undoManager isRedoing])
    {
        [undoManager setActionName: @"Insert Layer"];
    }

    [self postNotification_ReorderedLayers];
}

- (void) removeLayerAtIndex: (int) index
{
    int topLayerIndex;
    PPDocumentLayer *layer;
    NSData *archivedLayer = nil;
    bool needToSetDrawingLayer = NO;
    NSUndoManager *undoManager;

    if (![self hasLayerAtIndex: index])
    {
        return;
    }

    // manually invalidate relevant cached images that won't be invalidated later by
    // invalidateAllRelativeCachedLayerImagesForIndex: (called by handleUpdateToLayerAtIndex:)

    topLayerIndex = _numLayers - 1;
    [self invalidateCachedLayerImagesAtIndex: topLayerIndex];
    [self invalidateImagesInCache: _cachedOverlayImages
            fromIndex: index
            toIndex: topLayerIndex];

    layer = [_layers objectAtIndex: index];
    [layer setDelegate: nil];

    archivedLayer = [NSKeyedArchiver archivedDataWithRootObject: layer];

    [_layers removeObjectAtIndex: index];
    _numLayers = [_layers count];

    if (index == _indexOfDrawingLayer)
    {
        needToSetDrawingLayer = YES;
    }

    undoManager = [self undoManager];

    // need to register insertArchivedLayer:... undo invocation before the call to
    // createNewLayer, because createNewLayer registers an undo invocation for
    // removeLayerAtIndex: (which might cause layer ordering & draw layer index issues on
    // undo if the old removed layer is inserted before the newly-created layer is removed)

    [[undoManager prepareWithInvocationTarget: self]
                                                insertArchivedLayer: archivedLayer
                                                atIndex: index
                                                andSetAsDrawingLayer: needToSetDrawingLayer];

    if (!_numLayers)
    {
        if (![undoManager isUndoing] && ![undoManager isRedoing])
        {
            [self createNewLayer];
        }
    }
    else if (needToSetDrawingLayer)
    {
        index = _indexOfDrawingLayer;

        if (index > 0)
        {
            index--;

            [self invalidateImagesInCache: _cachedOverlayImages
                    fromIndex: index
                    toIndex: index];
        }

        [self setupDrawingLayerWithLayerAtIndex: index andPostNotification: NO];
    }
    else if (_indexOfDrawingLayer > index)
    {
        _indexOfDrawingLayer--;
    }

    if (index >= _numLayers)
    {
        index--;
    }

    [self handleUpdateToLayerAtIndex: index inRect: _canvasFrame];

    if (![undoManager isUndoing] && ![undoManager isRedoing])
    {
        [undoManager setActionName: @"Delete Layer"];
    }

    [self postNotification_ReorderedLayers];
}

- (void) moveLayerAtIndex: (int) oldIndex
            toIndex: (int) newIndex
{
    PPDocumentLayer *layer;
    NSUndoManager *undoManager;

    if (![self hasLayerAtIndex: oldIndex]
        || ![self hasLayerAtIndex: newIndex]
        || (oldIndex == newIndex))
    {
        return;
    }

    // manually invalidate relevant cached images that won't be invalidated later by
    // invalidateAllRelativeCachedLayerImagesForIndex: (called by
    // handleUpdateToLayerAtIndex: newIndex)

    if (oldIndex < newIndex)
    {
        [self invalidateImagesInCache: _cachedUnderlayImages
                fromIndex: oldIndex + 1
                toIndex: newIndex];
    }
    else
    {
        [self invalidateImagesInCache: _cachedOverlayImages
                fromIndex: newIndex
                toIndex: oldIndex - 1];
    }

    layer = [[[_layers objectAtIndex: oldIndex] retain] autorelease];

    [_layers removeObjectAtIndex: oldIndex];
    [_layers insertObject: layer atIndex: newIndex];

    _indexOfDrawingLayer = [_layers indexOfObject: _drawingLayer];

    undoManager = [self undoManager];
    [[undoManager prepareWithInvocationTarget: self] moveLayerAtIndex: newIndex
                                                                toIndex: oldIndex];

    if (![undoManager isUndoing] && ![undoManager isRedoing])
    {
        [undoManager setActionName: @"Reorder Layers"];
    }

    [self handleUpdateToLayerAtIndex: newIndex inRect: _canvasFrame];

    [self postNotification_ReorderedLayers];
}

- (void) duplicateLayerAtIndex: (int) index
{
    PPDocumentLayer *dupeLayer;
    NSString *dupeLayerName;

    if (![self hasLayerAtIndex: index])
    {
        return;
    }

    dupeLayer = [[[_layers objectAtIndex: index] copy] autorelease];

    if (!dupeLayer)
        return;

    dupeLayerName = [self duplicateNameForLayerName: [dupeLayer name]];

    [dupeLayer setName: dupeLayerName];

    [self insertLayer: dupeLayer atIndex: index + 1 andSetAsDrawingLayer: YES];

    [[self undoManager] setActionName: @"Duplicate Layer"];
}

- (void) removeAllLayers
{
    [self removeAllCachedLayersImages];

    [_layers makeObjectsPerformSelector: @selector(setDelegate:) withObject: nil];
    [_layers removeAllObjects];

    _numLayers = 0;
    _indexOfDrawingLayer = -1;

    [self setupDrawingLayerMembers];
}

- (void) removeNontargetLayers
{
    NSMutableArray *targetLayers;
    int i;
    NSBitmapImageRep *selectionMask = nil;

    targetLayers = [NSMutableArray array];

    if (!targetLayers)
        goto ERROR;

    [self setupTargetLayerIndexesForOperationTarget: _layerOperationTarget];

    // don't allow all layers to be removed (must be at least one target layer remaining)
    if (!_numTargetLayerIndexes)
        goto ERROR;

    for(i=0; i<_numTargetLayerIndexes; i++)
    {
        [targetLayers addObject: [_layers objectAtIndex: _targetLayerIndexes[i]]];
    }

    if (_hasSelection)
    {
        selectionMask = [[_selectionMask copy] autorelease];
    }

    [self setLayers: targetLayers];

    if (selectionMask)
    {
        [self setSelectionMask: selectionMask];
    }

    [[self undoManager] setActionName: @"Remove Nontarget Layers"];

    return;

ERROR:
    return;
}

- (bool) setLayers: (NSArray *) newLayers
{
    NSData *archivedOldLayers;
    unsigned oldIndexOfDrawingLayer, newIndexOfDrawingLayer;
    NSUndoManager *undoManager;

    if (!newLayers)
        goto ERROR;

    // use copies of the new layers, not the originals
    newLayers = [[[NSArray alloc] initWithArray: newLayers copyItems: YES] autorelease];

    if (!newLayers)
        goto ERROR;

    archivedOldLayers = [NSKeyedArchiver archivedDataWithRootObject: _layers];
    oldIndexOfDrawingLayer = _indexOfDrawingLayer;

    if (_hasSelection)
    {
        [self deselectAll]; // stores old selection in undo stack
    }

    [self removeAllLayers];

    [_layers setArray: newLayers];

    [_layers makeObjectsPerformSelector: @selector(setDelegate:) withObject: self];

    _numLayers = [_layers count];

    if (![self resizeCanvasForCurrentLayers])
    {
        goto ERROR;
    }

    newIndexOfDrawingLayer = oldIndexOfDrawingLayer;

    if (![self hasLayerAtIndex: newIndexOfDrawingLayer])
    {
        newIndexOfDrawingLayer = 0;
    }

    [self setupDrawingLayerWithLayerAtIndex: newIndexOfDrawingLayer andPostNotification: NO];

    [self handleUpdateToLayerAtIndex: _indexOfDrawingLayer inRect: _canvasFrame];

    undoManager = [self undoManager];

    [[undoManager prepareWithInvocationTarget: self]
                                    setupDrawingLayerWithLayerAtIndex: oldIndexOfDrawingLayer];

    [[undoManager prepareWithInvocationTarget: self]
                                    setLayersWithArchivedLayersData: archivedOldLayers];

    [self postNotification_ReloadedDocument];

    return YES;

ERROR:
    return NO;
}

- (void) selectDrawingLayerAtIndex: (int) newDrawingLayerIndex
{
    int oldIndexOfDrawingLayer;
    NSUndoManager *undoManager;

    if (![self hasLayerAtIndex: newDrawingLayerIndex]
        || (newDrawingLayerIndex == _indexOfDrawingLayer))
    {
        return;
    }

    oldIndexOfDrawingLayer = _indexOfDrawingLayer;

    [self setupDrawingLayerWithLayerAtIndex: newDrawingLayerIndex];

    undoManager = [self undoManager];
    [[undoManager prepareWithInvocationTarget: self]
                                            selectDrawingLayerAtIndex: oldIndexOfDrawingLayer];

    if (![undoManager isUndoing] && ![undoManager isRedoing])
    {
        [undoManager setActionName: @"Switch Drawing Layer"];
    }
}

- (int) indexOfDrawingLayer
{
    return _indexOfDrawingLayer;
}

- (PPDocumentLayer *) drawingLayer
{
    return _drawingLayer;
}

- (void) beginMultilayerOperation
{
    _disallowUpdatesToMergedBitmap = YES;

    [[[self undoManager] prepareWithInvocationTarget: self] finishMultilayerOperation];
}

- (void) finishMultilayerOperation
{
    _disallowUpdatesToMergedBitmap = NO;

    [self removeAllCachedLayersImages];
    [self handleUpdateToLayerAtIndex: _indexOfDrawingLayer inRect: _canvasFrame];

    [[[self undoManager] prepareWithInvocationTarget: self] beginMultilayerOperation];

    [self postNotification_PerformedMultilayerOperation];
}

- (void) moveDrawingLayerUp
{
    if ((_indexOfDrawingLayer >= (_numLayers - 1)) || (_indexOfDrawingLayer < 0))
    {
        return;
    }

    [self moveLayerAtIndex: _indexOfDrawingLayer toIndex: _indexOfDrawingLayer+1];

    [[self undoManager] setActionName: @"Move Layer Up"];
}

- (void) moveDrawingLayerDown
{
    if (_indexOfDrawingLayer < 1)
    {
        return;
    }

    [self moveLayerAtIndex: _indexOfDrawingLayer toIndex: _indexOfDrawingLayer-1];

    [[self undoManager] setActionName: @"Move Layer Down"];
}

- (void) mergeDrawingLayerUp
{
    NSBitmapImageRep *mergedLayersBitmap;

    if ((_indexOfDrawingLayer >= (_numLayers - 1)) || (_indexOfDrawingLayer < 0))
    {
        return;
    }

    mergedLayersBitmap = [self mergedLayersBitmapFromIndex: _indexOfDrawingLayer
                                toIndex: _indexOfDrawingLayer + 1];

    if (!mergedLayersBitmap)
        return;

    [self copyBitmapToDrawingLayer: mergedLayersBitmap atPoint: NSZeroPoint];

    [_drawingLayer setOpacity: 1.0f];
    [_drawingLayer setEnabled: YES];

    [self removeLayerAtIndex: _indexOfDrawingLayer + 1];

    [[self undoManager] setActionName: @"Merge with Layer Above"];
}

- (void) mergeDrawingLayerDown
{
    NSBitmapImageRep *mergedLayersBitmap;

    if (_indexOfDrawingLayer < 1)
    {
        return;
    }

    mergedLayersBitmap = [self mergedLayersBitmapFromIndex: _indexOfDrawingLayer - 1
                                toIndex: _indexOfDrawingLayer];

    if (!mergedLayersBitmap)
        return;

    [self copyBitmapToDrawingLayer: mergedLayersBitmap atPoint: NSZeroPoint];

    [_drawingLayer setOpacity: 1.0f];
    [_drawingLayer setEnabled: YES];

    [self removeLayerAtIndex: _indexOfDrawingLayer - 1];

    [[self undoManager] setActionName: @"Merge with Layer Below"];
}

- (void) mergeAllLayers
{
    PPDocumentLayer *mergedLayer;
    NSBitmapImageRep *selectionMask = nil;

    mergedLayer = [[_drawingLayer copy] autorelease];

    if (!mergedLayer)
        return;

    [[mergedLayer bitmap] ppCopyFromBitmap: _mergedVisibleLayersBitmap
                            toPoint: NSZeroPoint];

    [mergedLayer setOpacity: 1.0f];
    [mergedLayer setEnabled: YES];

    if (_hasSelection)
    {
        selectionMask = [[_selectionMask copy] autorelease];
    }

    [self setLayers: [NSArray arrayWithObject: mergedLayer]];

    if (selectionMask)
    {
        [self setSelectionMask: selectionMask];
    }

    [[self undoManager] setActionName: @"Merge All Layers"];
}

- (void) setEnabledFlagForAllLayers: (bool) isEnabled
{
    int i;

    isEnabled = (isEnabled) ? YES : NO;

    [self beginMultilayerOperation];

    for (i=0; i<_numLayers; i++)
    {
        [(PPDocumentLayer *) [_layers objectAtIndex: i] setEnabled: isEnabled];
    }

    [self finishMultilayerOperation];

    [[self undoManager]
                setActionName: (isEnabled) ? @"Enable All Layers" : @"Disable All Layers"];
}

- (void) copyBitmap: (NSBitmapImageRep *) bitmap
            toLayerAtIndex: (int) index
            atPoint: (NSPoint) origin
{
    PPDocumentLayer *layer;
    NSSize bitmapSize;
    NSRect updateRect;
    NSBitmapImageRep *layerBitmap;
    NSData *undoBitmapTIFFData;

    if (index == _indexOfDrawingLayer)
    {
        [self copyBitmapToDrawingLayer: bitmap atPoint: origin];
        return;
    }

    if (!bitmap)
        goto ERROR;

    layer = [self layerAtIndex: index];

    if (!layer)
        goto ERROR;

    origin = PPGeometry_PointClippedToIntegerValues(origin);
    bitmapSize = [bitmap ppSizeInPixels];

    updateRect.origin = origin;
    updateRect.size = bitmapSize;

    updateRect = NSIntersectionRect(updateRect, _canvasFrame);

    if (NSIsEmptyRect(updateRect))
    {
        goto ERROR;
    }

    if (!NSEqualSizes(updateRect.size, bitmapSize))
    {
        NSRect croppingBounds;

        croppingBounds.origin = NSMakePoint(updateRect.origin.x - origin.x,
                                            updateRect.origin.y - origin.y);

        croppingBounds.size = updateRect.size;

        bitmap = [bitmap ppShallowDuplicateFromBounds: croppingBounds];

        if (!bitmap)
            goto ERROR;
    }

    layerBitmap = [layer bitmap];

    undoBitmapTIFFData = [layerBitmap ppCompressedTIFFDataFromBounds: updateRect];

    if (!undoBitmapTIFFData)
        goto ERROR;

    [layerBitmap ppCopyFromBitmap: bitmap toPoint: updateRect.origin];
    [layer recacheImageInBounds: updateRect];

    [self handleUpdateToLayerAtIndex: index inRect: updateRect];

    [[[self undoManager] prepareWithInvocationTarget: self] copyTIFFData: undoBitmapTIFFData
                                                                toLayerAtIndex: index
                                                                atPoint: updateRect.origin];

    return;

ERROR:
    return;
}

- (void) handleUpdateToLayerAtIndex: (int) index inRect: (NSRect) updateRect
{
    if (![self hasLayerAtIndex: index] || _disallowUpdatesToMergedBitmap)
    {
        return;
    }

    updateRect = NSIntersectionRect(updateRect, _canvasFrame);

    if (NSIsEmptyRect(updateRect))
    {
        return;
    }

    [self invalidateAllRelativeCachedLayerImagesForIndex: index];

    [self updateMergedVisibleLayersBitmapInRect: updateRect
            indexOfUpdatedLayer: index];

    if (index == _indexOfDrawingLayer)
    {
        [self handleUpdateToDrawingLayerInRect: updateRect];
    }

    [self postNotification_UpdatedMergedVisibleAreaInRect: updateRect];

    if (!_disallowThumbnailImageUpdateNotifications)
    {
        [self postNotification_UpdatedMergedVisibleThumbnailImage];
    }
}

#pragma mark PPDocumentLayer delegate methods

- (void) layer: (PPDocumentLayer *) layer
            changedNameFromOldValue: (NSString *) oldName
{
    int layerIndex;
    NSUndoManager *undoManager;

    layerIndex = [self indexOfLayer: layer];

    if (layerIndex < 0)
    {
        return;
    }

    undoManager = [self undoManager];
    [[undoManager prepareWithInvocationTarget: self] setName: oldName
                                                        forLayerAtIndex: layerIndex];

    if (![undoManager isUndoing] && ![undoManager isRedoing])
    {
        [undoManager setActionName: @"Rename Layer"];
    }

    [self postNotification_ChangedAttributeOfLayerAtIndex: layerIndex];
}

- (void) layer: (PPDocumentLayer *) layer
            changedEnabledFlagFromOldValue: (bool) oldEnabledFlag
{
    int layerIndex;
    NSUndoManager *undoManager;

    if (![self handleUpdateToLayer: layer])
    {
        return;
    }

    layerIndex = [self indexOfLayer: layer];

    undoManager = [self undoManager];
    [[undoManager prepareWithInvocationTarget: self] setEnabledFlag: oldEnabledFlag
                                                        forLayerAtIndex: layerIndex];

    if (![undoManager isUndoing] && ![undoManager isRedoing])
    {
        [undoManager setActionName: [layer isEnabled] ? @"Enable Layer" : @"Disable Layer"];
    }

    [self postNotification_ChangedAttributeOfLayerAtIndex: layerIndex];
}

- (void) layer: (PPDocumentLayer *) layer
            changedOpacityFromOldValue: (float) oldOpacity
            shouldRegisterUndo: (bool) shouldRegisterUndo
{
    int layerIndex;

    if (![self handleUpdateToLayer: layer])
    {
        return;
    }

    layerIndex = [self indexOfLayer: layer];

    if (shouldRegisterUndo)
    {
        NSUndoManager *undoManager = [self undoManager];

        [[undoManager prepareWithInvocationTarget: self] setOpacity: oldOpacity
                                                            forLayerAtIndex: layerIndex];

        if (![undoManager isUndoing] && ![undoManager isRedoing])
        {
            NSString *actionName;

            actionName = ([layer opacity] > oldOpacity) ? @"Increase" : @"Decrease";
            actionName = [actionName stringByAppendingString: @" Layer Opacity"];

            if (!actionName)
            {
                actionName = @"Change Layer Opacity";
            }

            [[self undoManager] setActionName: actionName];
        }
    }

    [self postNotification_ChangedAttributeOfLayerAtIndex: layerIndex];
}

#pragma mark Private methods

- (bool) hasLayerAtIndex: (int) index
{
    return ((index >= 0) && (index < _numLayers)) ? YES : NO;
}

- (bool) isValidLayerInsertionIndex: (int) index
{
    return ((index >= 0)
            && (index <= _numLayers)
            && (_numLayers < kMaxLayersPerDocument)) ? YES : NO;
}

- (int) indexOfLayer: (PPDocumentLayer *) layer
{
    NSUInteger indexOfLayer;

    indexOfLayer = [_layers indexOfObject: layer];

    if (indexOfLayer == NSNotFound)
    {
        return -1;
    }

    return (int) indexOfLayer;
}

- (void) invalidateAllRelativeCachedLayerImagesForIndex: (int) index
{
    if (![self hasLayerAtIndex: index])
    {
        return;
    }

    [self invalidateImagesInCache: _cachedOverlayImages
            fromIndex: 0
            toIndex: index - 1];

    [self invalidateImagesInCache: _cachedUnderlayImages
            fromIndex: index + 1
            toIndex: _numLayers - 1];
}

- (void) invalidateCachedLayerImagesAtIndex: (int) index
{
    if (![self hasLayerAtIndex: index])
    {
        return;
    }

    [self invalidateImagesInCache: _cachedOverlayImages
            fromIndex: index
            toIndex: index];

    [self invalidateImagesInCache: _cachedUnderlayImages
            fromIndex: index
            toIndex: index];
}

- (void) invalidateImagesInCache: (NSImage **) cachedImagesArray
            fromIndex: (int) startIndex
            toIndex: (int) endIndex
{
    int i;

    if (startIndex < 0)
    {
        startIndex = 0;
    }

    if (endIndex >= _numLayers)
    {
        endIndex = _numLayers - 1;
    }

    if (startIndex > endIndex)
    {
        return;
    }

    for (i=startIndex; i<=endIndex; i++)
    {
        if (cachedImagesArray[i])
        {
            [cachedImagesArray[i] release];
            cachedImagesArray[i] = nil;
        }
    }
}

- (void) removeAllCachedLayersImages
{
    [self invalidateImagesInCache: _cachedOverlayImages
            fromIndex: 0
            toIndex: _numLayers - 1];

    [self invalidateImagesInCache: _cachedUnderlayImages
            fromIndex: 0
            toIndex: _numLayers - 1];
}

- (NSImage *) cachedOverlayersImageForIndex: (int) index
{
    if ((index >= (_numLayers - 1)) || ![self hasLayerAtIndex: index])
    {
        return nil;
    }

    if (!_cachedOverlayImages[index])
    {
        _cachedOverlayImages[index] =
                [[self mergedLayersImageFromIndex: index + 1 toIndex: _numLayers - 1] retain];
    }

    return _cachedOverlayImages[index];
}

- (NSImage *) cachedUnderlayersImageForIndex: (int) index
{
    if ((index < 1) || ![self hasLayerAtIndex: index])
    {
        return nil;
    }

    if (!_cachedUnderlayImages[index])
    {
        _cachedUnderlayImages[index] =
                            [[self mergedLayersImageFromIndex: 0 toIndex: index - 1] retain];
    }

    return _cachedUnderlayImages[index];
}

- (NSImage *) mergedLayersImageFromIndex: (int) firstIndex
                toIndex: (int) lastIndex
{
    NSBitmapImageRep *mergedLayersBitmap;

    if (firstIndex < 0)
    {
        firstIndex = 0;
    }

    if (lastIndex >= [_layers count])
    {
        lastIndex = [_layers count] - 1;
    }

    if (firstIndex > lastIndex)
    {
        goto ERROR;
    }

    mergedLayersBitmap = [self mergedLayersBitmapFromIndex: firstIndex toIndex: lastIndex];

    if (!mergedLayersBitmap)
        goto ERROR;

    return [NSImage ppImageWithBitmap: mergedLayersBitmap];

ERROR:
    return nil;
}

- (NSBitmapImageRep *) mergedLayersBitmapFromIndex: (int) firstIndex
                        toIndex: (int) lastIndex
{
    NSBitmapImageRep *mergedLayersBitmap;
    int i;
    PPDocumentLayer *layer;
    float layerOpacity;

    if (firstIndex <= 0)
    {
        firstIndex = 0;
    }

    if (lastIndex >= [_layers count])
    {
        lastIndex = [_layers count] - 1;
    }

    if (firstIndex > lastIndex)
    {
        goto ERROR;
    }

    mergedLayersBitmap = [NSBitmapImageRep ppImageBitmapOfSize: _canvasFrame.size];

    if (!mergedLayersBitmap)
        goto ERROR;

    [mergedLayersBitmap ppSetAsCurrentGraphicsContext];

    for (i=firstIndex; i<=lastIndex; i++)
    {
        layer = [self layerAtIndex: i];
        layerOpacity = [layer opacity];

        if ([layer isEnabled] && (layerOpacity > 0.0f))
        {
            [[layer image] drawInRect: _canvasFrame
                            fromRect: _canvasFrame
                            operation: NSCompositeSourceOver
                            fraction: layerOpacity];
        }
    }

    [mergedLayersBitmap ppRestoreGraphicsContext];

    return mergedLayersBitmap;

ERROR:
    return nil;
}

- (void) updateMergedVisibleLayersBitmapInRect: (NSRect) rect
            indexOfUpdatedLayer: (int) indexOfUpdatedLayer
{
    NSImage *mergedUnderlayersImage, *mergedOverlayersImage, *updatedLayerImage;
    PPDocumentLayer *updatedLayer;
    float updatedLayerOpacity;

    if (NSIsEmptyRect(rect)
        || ![self hasLayerAtIndex: indexOfUpdatedLayer])
    {
        return;
    }

    mergedUnderlayersImage = [self cachedUnderlayersImageForIndex: indexOfUpdatedLayer];
    mergedOverlayersImage = [self cachedOverlayersImageForIndex: indexOfUpdatedLayer];

    updatedLayer = [_layers objectAtIndex: indexOfUpdatedLayer];
    updatedLayerOpacity = [updatedLayer opacity];

    if ([updatedLayer isEnabled] && (updatedLayerOpacity > 0.0f))
    {
        updatedLayerImage = [updatedLayer image];
    }
    else
    {
        updatedLayerImage = nil;
    }

    [_mergedVisibleLayersBitmap ppSetAsCurrentGraphicsContext];

    if (mergedUnderlayersImage)
    {
        [mergedUnderlayersImage drawInRect: rect
                                fromRect: rect
                                operation: NSCompositeCopy
                                fraction: 1.0f];
    }
    else
    {
        [_mergedVisibleLayersBitmap ppClearBitmapInBounds: rect];
    }

    if (updatedLayerImage)
    {
        [updatedLayerImage drawInRect: rect
                            fromRect: rect
                            operation: NSCompositeSourceOver
                            fraction: updatedLayerOpacity];
    }

    if (mergedOverlayersImage)
    {
        [mergedOverlayersImage drawInRect: rect
                                fromRect: rect
                                operation: NSCompositeSourceOver
                                fraction: 1.0f];
    }

    [_mergedVisibleLayersBitmap ppRestoreGraphicsContext];

    [self recacheMergedVisibleLayersThumbnailImageInBounds: rect];
}

- (NSString *) uniqueLayerNameWithRoot: (NSString *) rootName
{
    int rootNameLength;
    NSMutableSet *prefixMatches;
    NSEnumerator *enumerator;
    PPDocumentLayer *layer;
    NSString *layerName, *testName;
    int suffixValue;

    if (![rootName length])
    {
        rootName = @"";
    }

    rootNameLength = [rootName length];

    prefixMatches = [NSMutableSet set];

    enumerator = [_layers objectEnumerator];

    while (layer = [enumerator nextObject])
    {
        layerName = [layer name];

        if (([layerName length] >= rootNameLength)
            && [layerName hasPrefix: rootName])
        {
            [prefixMatches addObject: layerName];
        }
    }

    if (![prefixMatches count])
    {
        return rootName;
    }

    suffixValue = 1;
    testName = [NSString stringWithFormat: @"%@ %02d", rootName, suffixValue];

    while ([prefixMatches containsObject: testName])
    {
        suffixValue++;
        testName = [NSString stringWithFormat: @"%@ %02d", rootName, suffixValue];
    }

    return testName;
}

- (NSString *) duplicateNameForLayerName: (NSString *) layerName
{
    if (![layerName hasSuffix: @" copy"])
    {
        layerName = [layerName stringByAppendingString: @" copy"];
    }

    return [self uniqueLayerNameWithRoot: layerName];
}

- (void) setupDrawingLayerWithLayerAtIndex: (int) newDrawingLayerIndex
            andPostNotification: (bool) shouldPostNotification
{
    if (![self hasLayerAtIndex: newDrawingLayerIndex])
    {
        newDrawingLayerIndex = -1;
    }

    _indexOfDrawingLayer = newDrawingLayerIndex;

    [self setupDrawingLayerMembers];

    [self handleUpdateToDrawingLayerInRect: _canvasFrame];

    if (shouldPostNotification)
    {
        [self postNotification_SwitchedDrawingLayer];
    }
}

- (void) setupDrawingLayerWithLayerAtIndex: (int) newDrawingLayerIndex
{
    [self setupDrawingLayerWithLayerAtIndex: newDrawingLayerIndex andPostNotification: YES];
}

- (void) setupDrawingLayerMembers
{
    [_drawingLayerImage release];
    _drawingLayerImage = nil;

    [_drawingLayerBitmap release];
    _drawingLayerBitmap = nil;

    [_drawingLayer release];
    _drawingLayer = nil;

    if ([self hasLayerAtIndex: _indexOfDrawingLayer])
    {
        _drawingLayer = [[_layers objectAtIndex: _indexOfDrawingLayer] retain];
        _drawingLayerBitmap = [[_drawingLayer bitmap] retain];
        _drawingLayerImage = [[_drawingLayer image] retain];

        [_drawingUndoBitmap ppCopyFromBitmap: _drawingLayerBitmap toPoint: NSZeroPoint];
    }
    else
    {
        [_drawingUndoBitmap ppClearBitmap];
    }
}

- (bool) handleUpdateToLayer: (PPDocumentLayer *) layer
{
    int indexOfChangedLayer;

    indexOfChangedLayer = [_layers indexOfObject: layer];

    if (![self hasLayerAtIndex: indexOfChangedLayer])
    {
        return NO;
    }

    [self handleUpdateToLayerAtIndex: indexOfChangedLayer inRect: _canvasFrame];

    return YES;
}

- (void) handleUpdateToDrawingLayerInRect: (NSRect) updateRect
{
    [self updateDissolvedDrawingLayerBitmapInRect: updateRect];

    [self postNotification_UpdatedDrawingLayerAreaInRect: updateRect];

    if (!_disallowThumbnailImageUpdateNotifications)
    {
        [self postNotification_UpdatedDrawingLayerThumbnailImage];
    }
}

- (void) updateDissolvedDrawingLayerBitmapInRect: (NSRect) updateRect
{
    [_dissolvedDrawingLayerBitmap ppClearBitmapInBounds: updateRect];

    if ([_drawingLayer isEnabled])
    {
        [_dissolvedDrawingLayerBitmap ppSetAsCurrentGraphicsContext];

        [_drawingLayerImage drawInRect: updateRect
                            fromRect: updateRect
                            operation: NSCompositeSourceOver
                            fraction: [_drawingLayer opacity]];

        [_dissolvedDrawingLayerBitmap ppRestoreGraphicsContext];
    }

    [self recacheDissolvedDrawingLayerThumbnailImageInBounds: updateRect];
}

- (void) insertArchivedLayer: (NSData *) archivedLayer
            atIndex: (int) index
            andSetAsDrawingLayer: (bool) shouldSetAsDrawingLayer
{
    PPDocumentLayer *layer;

    if (!archivedLayer)
        goto ERROR;

    layer = [NSKeyedUnarchiver unarchiveObjectWithData: archivedLayer];

    if (!layer || ![layer isKindOfClass: [PPDocumentLayer class]])
    {
        goto ERROR;
    }

    [self insertLayer: layer atIndex: index andSetAsDrawingLayer: shouldSetAsDrawingLayer];

    return;

ERROR:
    return;
}

- (bool) setLayersWithArchivedLayersData: (NSData *) archivedLayersData
{
    NSArray *layers;

    if (!archivedLayersData)
        goto ERROR;

    layers = [NSKeyedUnarchiver unarchiveObjectWithData: archivedLayersData];

    if (!layers || ![layers isKindOfClass: [NSArray class]])
    {
        goto ERROR;
    }

    return [self setLayers: layers];

ERROR:
    return NO;
}

- (void) setName: (NSString *) name forLayerAtIndex: (int) index
{
    [[self layerAtIndex: index] setName: name];
}

- (void) setEnabledFlag: (bool) isEnabled forLayerAtIndex: (int) index
{
    [[self layerAtIndex: index] setEnabled: isEnabled];
}

- (void) setOpacity: (float) opacity forLayerAtIndex: (int) index
{
    [[self layerAtIndex: index] setOpacity: opacity];
}

- (void) copyTIFFData: (NSData *) tiffData
            toLayerAtIndex: (int) index
            atPoint: (NSPoint) origin
{
    [self copyBitmap: [NSBitmapImageRep imageRepWithData: tiffData]
            toLayerAtIndex: index
            atPoint: origin];
}

// recacheMergedVisibleLayersThumbnailImageInBounds:
// & recacheDissolvedDrawingLayerThumbnailImageInBounds: methods are patch targets on GNUstep
// (PPGNUstepGlue_ImageRecacheSpeedups)

- (void) recacheMergedVisibleLayersThumbnailImageInBounds: (NSRect) bounds
{
    [_mergedVisibleLayersThumbnailImage recache];
}

- (void) recacheDissolvedDrawingLayerThumbnailImageInBounds: (NSRect) bounds
{
    [_dissolvedDrawingLayerThumbnailImage recache];
}

@end
